/*
  Liquid War 6 is a unique multiplayer wargame.
  Copyright (C)  2005, 2006, 2007, 2008, 2009  Christian Mauduit <ufoot@ufoot.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  

  Liquid War 6 homepage : http://www.gnu.org/software/liquidwar6/
  Contact author        : ufoot@ufoot.org
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <math.h>

#include "ker.h"
#include "ker-internal.h"

/*
 * The code in this file is heavily inspired from Liquid War 5,
 * it's an adaptation of "src/mesh.c".
 */

/*
 * Internal structure used just when creating the mesh.
 * The idea is to have w*h zones, of size 1x1. Of course
 * these will be grouped afterwards, but when creating the
 * zones, it's usefull to have this in memory.
 * This was called "MESHER" in Liquid War 5.
 */
typedef struct draft_zone_s
{
  lw6sys_xyz_t pos;
  int32_t used;
  int32_t size;
  int32_t link[LW6KER_NB_DIRS];
  int32_t corres;
}
draft_zone_t;

typedef struct draft_zones_s
{
  lw6sys_whd_t shape;
  int32_t max_zone_size;
  draft_zone_t *zones;
}
draft_zones_t;

void
_lw6ker_map_struct_update_checksum (lw6ker_map_struct_t *
				    map_struct, u_int32_t * checksum)
{
  int i;

  lw6sys_checksum_update_whd (checksum, &(map_struct->shape));
  lw6sys_checksum_update_int32 (checksum, map_struct->shape_surface);
  lw6sys_checksum_update_int32 (checksum, map_struct->nb_zones);
  lw6sys_checksum_update_int32 (checksum, map_struct->nb_slots);
  lw6sys_checksum_update_int32 (checksum, map_struct->nb_usable_slots);
  lw6sys_checksum_update_int32 (checksum, map_struct->room_for_armies);
  lw6sys_checksum_update_int32 (checksum, map_struct->max_zone_size);
  for (i = 0; i < map_struct->nb_zones; ++i)
    {
      _lw6ker_zone_struct_update_checksum (&(map_struct->zones[i]), checksum);
    }
  for (i = 0; i < map_struct->nb_slots; ++i)
    {
      _lw6ker_slot_struct_update_checksum (&(map_struct->slots[i]), checksum);
    }
}

int
_lw6ker_map_struct_lazy_compare (lw6ker_map_struct_t *
				 map_struct_a, lw6ker_map_struct_t *
				 map_struct_b)
{
  int ret = 0;

  if (map_struct_a && map_struct_b)
    {
      if (map_struct_a == map_struct_b)
	{
	  ret = 1;
	}
      else
	{
	  ret = (map_struct_a->shape.w == map_struct_b->shape.w) &&
	    (map_struct_a->shape.h == map_struct_b->shape.h) &&
	    (map_struct_a->shape.d == map_struct_b->shape.d) &&
	    (map_struct_a->shape_surface == map_struct_b->shape_surface) &&
	    (map_struct_a->nb_zones == map_struct_b->nb_zones) &&
	    (map_struct_a->nb_slots == map_struct_b->nb_slots) &&
	    (map_struct_a->nb_usable_slots == map_struct_b->nb_usable_slots)
	    && (map_struct_a->room_for_armies ==
		map_struct_b->room_for_armies)
	    && (map_struct_a->max_zone_size == map_struct_b->max_zone_size);
	}
    }

  if (!ret)
    {
      lw6sys_log (LW6SYS_LOG_DEBUG, _("map_structs do not look the same"));
    }

  return ret;
}

static int32_t
relative_index_no_check (lw6map_level_t * level, int32_t x, int32_t y,
			 int32_t z)
{
  return z * level->body.shape.w * level->body.shape.h +
    y * level->body.shape.w + x;
}

static int32_t
relative_index (lw6map_level_t * level, int32_t x, int32_t y, int32_t z)
{
  lw6map_coords_fix_xy (&(level->param.rules), &(level->body.shape), &x, &y);
  lw6map_coords_fix_z (&(level->param.rules), &(level->body.shape), &z);

  return relative_index_no_check (level, x, y, z);
}

/*
 * Creates a new draft_zones_t object. The idea is to create a structure which
 * is of the same nature as the final MAP_STRUCT "zones", but unoptimized.
 * We create a structure (a "MESH" in Liquid War 5)) where each square
 * has pointers on its neighbours. As there are 12 directions, these is
 * very redundant. This is just a temporary struct which will be
 * optimized afterwards.
 */
static draft_zones_t *
draft_zones_new (lw6map_level_t * level, lw6sys_progress_t * progress)
{
  draft_zones_t *ret;

  lw6sys_progress_begin (progress);

  ret = (draft_zones_t *) LW6SYS_CALLOC (sizeof (draft_zones_t));
  if (ret)
    {
      int32_t w, h, d, size;

      ret->shape = level->body.shape;
      ret->max_zone_size = 1;

      w = ret->shape.w;
      h = ret->shape.h;
      d = ret->shape.d;
      size = w * h * d;

      if (size >= LW6MAP_MIN_SURFACE)
	{
	  ret->zones =
	    (draft_zone_t *) LW6SYS_CALLOC (size * sizeof (draft_zone_t));

	  if (ret->zones)
	    {
	      int32_t i, j, x, y, z;

	      for (i = 0; i < size; ++i)
		{
		  ret->zones[i].size = 1;
		  for (j = 0; j < LW6KER_NB_DIRS; ++j)
		    {
		      ret->zones[i].link[j] = -1;
		    }
		}

	      for (z = 0; z < d; ++z)
		{
		  for (y = 0; y < h; ++y)
		    {
		      for (x = 0; x < w; ++x)
			{
			  i = relative_index_no_check (level, x, y, z);

			  ret->zones[i].pos.x = x;
			  ret->zones[i].pos.y = y;
			  ret->zones[i].pos.z = z;
			  ret->zones[i].used =
			    lw6map_body_get (&(level->body), x, y,
					     z) > 0 ? 1 : 0;
			}
		    }
		}

	      for (z = 0; z < d; ++z)
		{
		  for (y = 0; y < h; ++y)
		    {
		      for (x = 0; x < w; ++x)
			{
			  i = relative_index (level, x, y, z);
			  if (ret->zones[i].used)
			    {
			      j = relative_index (level, x, y - 1, z);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_NNW] =
				    ret->zones[i].link[LW6KER_DIR_NNE] = j;
				}
			      j = relative_index (level, x + 1, y - 1, z);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_NE] = j;
				}
			      j = relative_index (level, x + 1, y, z);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_ENE] =
				    ret->zones[i].link[LW6KER_DIR_ESE] = j;
				}
			      j = relative_index (level, x + 1, y + 1, z);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_SE] = j;
				}
			      j = relative_index (level, x, y + 1, z);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_SSE] =
				    ret->zones[i].link[LW6KER_DIR_SSW] = j;
				}
			      j = relative_index (level, x - 1, y + 1, z);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_SW] = j;
				}
			      j = relative_index (level, x - 1, y, z);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_WSW] =
				    ret->zones[i].link[LW6KER_DIR_WNW] = j;
				}
			      j = relative_index (level, x - 1, y - 1, z);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_NW] = j;
				}
			      j = relative_index (level, x, y, z - 1);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_UP] = j;
				}
			      j = relative_index (level, x, y, z + 1);
			      if (ret->zones[j].used && i != j)
				{
				  ret->zones[i].link[LW6KER_DIR_DOWN] = j;
				}
			    }
			}
		    }
		}
	    }
	  else
	    {
	      lw6sys_log (LW6SYS_LOG_WARNING,
			  _("unable to allocate %d draft zones"), size);
	    }
	}
      else
	{
	  lw6sys_log (LW6SYS_LOG_WARNING,
		      _("map is too small, size=%d, LW6MAP_MIN_SURFACE=%d"),
		      size, LW6MAP_MIN_SURFACE);
	}

    }

  lw6sys_progress_end (progress);

  return ret;
}

static void
draft_zones_free (draft_zones_t * draft_zones)
{
  if (draft_zones)
    {
      if (draft_zones->zones)
	{
	  LW6SYS_FREE (draft_zones->zones);
	}
      LW6SYS_FREE (draft_zones);
    }
}

/*
 * Here we scan the whole draft_zones_t data, and whenever we find
 * 4 zones which form a square and respect "some conditions",
 * we group them into a bigger zone. The only thing to be carefull
 * about is to handle correctly all the pointers (links) on
 * other zones, especially those that link back to the squares
 * we suppress.
 */
static int32_t
draft_zones_group (lw6map_level_t * level, draft_zones_t * draft_zones,
		   int32_t step)
{
  int32_t found = 0;
  int32_t x, y, z, w, h, d, size, i, j, k;
  int32_t i_ne, i_se, i_sw, i_nw, i_up, i_down;
  draft_zone_t *zone_ne, *zone_se, *zone_sw, *zone_nw, *zone_up, *zone_down;
  int32_t *test;

  w = draft_zones->shape.w;
  h = draft_zones->shape.h;
  d = draft_zones->shape.d;
  size = d * h * w;

  for (z = 0; z < d; z++)
    {
      for (y = 0; y < h - step; y += step * 2)
	{
	  for (x = 0; x < w - step; x += step * 2)
	    {
	      i = z * w * h + y * w + x;
	      if (draft_zones->zones[i].used)
		{
		  i_ne = i + step;
		  i_se = i + (w + 1) * step;
		  i_sw = i + w * step;
		  i_nw = i;
		  zone_ne = &draft_zones->zones[i_ne];
		  zone_se = &draft_zones->zones[i_se];
		  zone_sw = &draft_zones->zones[i_sw];
		  zone_nw = &draft_zones->zones[i_nw];

		  i_up = i - w * h;	// may be out of bounds
		  if (i_up >= 0 && i_up < size)
		    {
		      zone_up = &draft_zones->zones[i_up];
		    }
		  else
		    {
		      zone_up = NULL;
		    }
		  i_down = i + w * h;	// may be out of bounds
		  if (i_down >= 0 && i_down < size)
		    {
		      zone_down = &draft_zones->zones[i_down];
		    }
		  else
		    {
		      zone_down = NULL;
		    }

		  /*
		   * This test contains all the criteria which make
		   * us decide wether we group the squares are not.
		   * It's a wise choice to be rather restrictive.
		   * For instance the fact tha NE SW NW NE links
		   * actually point to something might seem useless.
		   * It's not, the main reason is that it's a rather
		   * good choice that whenever there's an obstacle,
		   * the zones around it are forced to be small.
		   */
		  if (zone_ne->used && zone_ne->size == step
		      && zone_se->used && zone_se->size == step
		      && zone_sw->used && zone_sw->size == step
		      && zone_nw->used && zone_nw->size == step
		      && zone_ne->link[LW6KER_DIR_NNW] ==
		      zone_ne->link[LW6KER_DIR_NNE]
		      && zone_ne->link[LW6KER_DIR_ENE] ==
		      zone_ne->link[LW6KER_DIR_ESE]
		      && zone_se->link[LW6KER_DIR_ENE] ==
		      zone_se->link[LW6KER_DIR_ESE]
		      && zone_se->link[LW6KER_DIR_SSE] ==
		      zone_se->link[LW6KER_DIR_SSW]
		      && zone_sw->link[LW6KER_DIR_SSE] ==
		      zone_sw->link[LW6KER_DIR_SSW]
		      && zone_sw->link[LW6KER_DIR_WSW] ==
		      zone_sw->link[LW6KER_DIR_WNW]
		      && zone_nw->link[LW6KER_DIR_WSW] ==
		      zone_nw->link[LW6KER_DIR_WNW]
		      && zone_nw->link[LW6KER_DIR_NNW] ==
		      zone_nw->link[LW6KER_DIR_NNE]
		      && zone_ne->link[LW6KER_DIR_NE] >= 0
		      && zone_se->link[LW6KER_DIR_SE] >= 0
		      && zone_sw->link[LW6KER_DIR_SW] >= 0
		      && zone_nw->link[LW6KER_DIR_NW] >= 0
		      && ((!zone_up) || (zone_up && zone_up->size >= step))
		      && ((!zone_down)
			  || (zone_down && zone_down->size >= step)))
		    {
		      /*
		       * Delete unused zones
		       */
		      zone_ne->used = 0;
		      zone_se->used = 0;
		      zone_sw->used = 0;

		      /*
		       * Double the zone size of the remaining zone
		       */
		      zone_nw->size = step * 2;

		      /*
		       * Change the links of the remaining zone
		       */
		      zone_nw->link[LW6KER_DIR_NNE] =
			zone_ne->link[LW6KER_DIR_NNE];
		      zone_nw->link[LW6KER_DIR_NE] =
			zone_ne->link[LW6KER_DIR_NE];
		      zone_nw->link[LW6KER_DIR_ENE] =
			zone_ne->link[LW6KER_DIR_ENE];
		      zone_nw->link[LW6KER_DIR_ESE] =
			zone_se->link[LW6KER_DIR_ESE];
		      zone_nw->link[LW6KER_DIR_SE] =
			zone_se->link[LW6KER_DIR_SE];
		      zone_nw->link[LW6KER_DIR_SSE] =
			zone_se->link[LW6KER_DIR_SSE];
		      zone_nw->link[LW6KER_DIR_SSW] =
			zone_sw->link[LW6KER_DIR_SSW];
		      zone_nw->link[LW6KER_DIR_SW] =
			zone_sw->link[LW6KER_DIR_SW];
		      zone_nw->link[LW6KER_DIR_WSW] =
			zone_sw->link[LW6KER_DIR_WSW];

		      /*
		       * Change all the links of the adjacent zones,
		       * which were pointing on the deleted zones,
		       * and must now point on the remaining one.
		       */
		      for (j = 0; j < LW6KER_DIR_UP; ++j)
			{
			  for (k = 0; k < LW6KER_DIR_UP; ++k)
			    {
			      if (zone_nw->link[j] >= 0)
				{
				  test =
				    &(draft_zones->zones[zone_nw->link[j]].
				      link[k]);
				  if (*test == i_ne || *test == i_se
				      || *test == i_sw)
				    {
				      *test = i_nw;
				    }
				}
			    }
			}

		      /*
		       * Now treat vertical (cross-layer) zones links
		       * which were pointing on the deleted zones
		       */
		      if (d > 1)
			{
			  j =
			    relative_index (level,
					    draft_zones->zones[i_ne].pos.x,
					    draft_zones->zones[i_ne].pos.y,
					    draft_zones->zones[i_ne].pos.z +
					    1);
			  if (draft_zones->zones[j].used)
			    {
			      draft_zones->zones[j].link[LW6KER_DIR_UP] =
				i_nw;
			    }
			  j =
			    relative_index (level,
					    draft_zones->zones[i_se].pos.x,
					    draft_zones->zones[i_se].pos.y,
					    draft_zones->zones[i_se].pos.z +
					    1);
			  if (draft_zones->zones[j].used)
			    {
			      draft_zones->zones[j].link[LW6KER_DIR_UP] =
				i_nw;
			    }
			  j =
			    relative_index (level,
					    draft_zones->zones[i_sw].pos.x,
					    draft_zones->zones[i_sw].pos.y,
					    draft_zones->zones[i_sw].pos.z +
					    1);
			  if (draft_zones->zones[j].used)
			    {
			      draft_zones->zones[j].link[LW6KER_DIR_UP] =
				i_nw;
			    }
			  j =
			    relative_index (level,
					    draft_zones->zones[i_ne].pos.x,
					    draft_zones->zones[i_ne].pos.y,
					    draft_zones->zones[i_ne].pos.z -
					    1);
			  if (draft_zones->zones[j].used)
			    {
			      draft_zones->zones[j].link[LW6KER_DIR_DOWN] =
				i_nw;
			    }
			  j =
			    relative_index (level,
					    draft_zones->zones[i_se].pos.x,
					    draft_zones->zones[i_se].pos.y,
					    draft_zones->zones[i_se].pos.z -
					    1);
			  if (draft_zones->zones[j].used)
			    {
			      draft_zones->zones[j].link[LW6KER_DIR_DOWN] =
				i_nw;
			    }
			  j =
			    relative_index (level,
					    draft_zones->zones[i_sw].pos.x,
					    draft_zones->zones[i_sw].pos.y,
					    draft_zones->zones[i_sw].pos.z -
					    1);
			  if (draft_zones->zones[j].used)
			    {
			      draft_zones->zones[j].link[LW6KER_DIR_DOWN] =
				i_nw;
			    }
			}

		      draft_zones->max_zone_size = zone_nw->size;
		      ++found;
		    }
		}
	    }
	}
    }

  return found;
}

/*
 * Transform the temporary draft_zones_t into the proper
 * final MAP_STRUCT. The major change is to actually
 * delete and remove from memory all the unused ZONES.
 * We "compact" the zones so that instead of using
 * tons of RAM, we use just what we need. While this
 * complicates a little bit the algorithms (we can't
 * assume any more than y * w + x = i) reducing memory
 * usage is always a good thing, the more we reduce it,
 * the more we get a chance to have things cached.
 */
static int
draft_zones_to_map_struct (lw6ker_map_struct_t * map_struct,
			   draft_zones_t * draft_zones,
			   lw6sys_progress_t * progress)
{
  int ret = 0;
  int32_t w, h, d, size, nb_zones;
  int32_t i, j, k, temp;

  lw6sys_progress_begin (progress);

  w = draft_zones->shape.w;
  h = draft_zones->shape.h;
  d = draft_zones->shape.d;
  size = w * h * d;
  nb_zones = 0;

  for (i = 0; i < size; ++i)
    {
      if (draft_zones->zones[i].used)
	{
	  ++nb_zones;
	}
    }

  map_struct->shape = draft_zones->shape;
  map_struct->shape_surface = w * h;
  map_struct->nb_slots = w * h * d;
  map_struct->nb_zones = nb_zones;
  map_struct->max_zone_size = draft_zones->max_zone_size;

  map_struct->zones =
    (lw6ker_zone_struct_t *) LW6SYS_MALLOC (nb_zones *
					    sizeof (lw6ker_zone_struct_t));

  if (map_struct->zones)
    {
      j = 0;
      for (i = 0; i < size; ++i)
	{
	  if (draft_zones->zones[i].used)
	    {
	      map_struct->zones[j].pos = draft_zones->zones[i].pos;
	      map_struct->zones[j].size = draft_zones->zones[i].size;
	      for (k = 0; k < LW6KER_NB_DIRS; ++k)
		{
		  map_struct->zones[j].link[k] =
		    draft_zones->zones[i].link[k];
		}
	      draft_zones->zones[i].corres = j;

	      j++;
	    }
	  else
	    {
	      draft_zones->zones[i].corres = -1;
	    }
	}

      for (j = 0; j < nb_zones; ++j)
	{
	  for (k = 0; k < LW6KER_NB_DIRS; ++k)
	    {
	      if ((temp = (map_struct->zones[j].link[k])) >= 0)
		{
		  map_struct->zones[j].link[k] =
		    draft_zones->zones[temp].corres;
		}
	      else
		{
		  map_struct->zones[j].link[k] = -1;
		}
	    }
	}
      ret = 1;
    }
  else
    {
      lw6sys_log (LW6SYS_LOG_WARNING,
		  _("unable to create zones for map struct"));
    }

  lw6sys_progress_end (progress);

  return ret;
}

/*
 * Create the slots, the main interest in doing this is to
 * have a quick way to get the right zone for a given (x,y).
 */
static int
init_slots (lw6ker_map_struct_t * map_struct, lw6map_level_t * level,
	    lw6sys_progress_t * progress)
{
  int ret = 0;

  lw6sys_progress_begin (progress);

  map_struct->slots =
    (lw6ker_slot_struct_t *) LW6SYS_CALLOC (map_struct->shape.w *
					    map_struct->shape.h *
					    map_struct->shape.d *
					    sizeof (lw6ker_slot_struct_t));

  if (map_struct)
    {
      int32_t i, x, y, z;

      lw6ker_zone_struct_t *zone;

      /*
       * Set up the id lookup table
       */
      for (z = 0; z < map_struct->shape.d; ++z)
	{
	  for (y = 0; y < map_struct->shape.h; ++y)
	    {
	      for (x = 0; x < map_struct->shape.w; ++x)
		{
		  lw6ker_map_struct_set_zone_id (map_struct, x, y, z, -1);
		}
	    }
	}

      for (i = 0; i < map_struct->nb_zones; ++i)
	{
	  zone = &map_struct->zones[i];
	  for (y = 0; y < map_struct->zones[i].size; ++y)
	    {
	      for (x = 0; x < map_struct->zones[i].size; ++x)
		{
		  lw6ker_map_struct_set_zone_id (map_struct, zone->pos.x + x,
						 zone->pos.y + y, zone->pos.z,
						 i);
		  map_struct->nb_usable_slots++;
		  map_struct->room_for_armies++;
		}
	    }
	}

      ret = 1;
    }
  else
    {
      lw6sys_log (LW6SYS_LOG_WARNING,
		  _("unable to create slots for map struct"));
    }

  lw6sys_progress_end (progress);

  return ret;
}

/*
 * Returns a "useless" ratio which is "how much we compressed
 * the map by using all this ZONE wizardry. Just to keep track
 * of wether it's usefull to implement this. Debugging tool.
 */
float
_lw6ker_map_struct_get_compression (lw6ker_map_struct_t * map_struct)
{
  float ret = 0.0f;

  if (map_struct->nb_slots > 0)
    {
      ret = ((float) map_struct->nb_zones) / ((float) (map_struct->nb_slots));
    }

  return ret;
}

/*
 * This is one key function: it calls the transformation of a lw6map_level_t,
 * that is a plain object which reflects only the content of PNG & XML
 * files, into an in-memory "ready for algorithmic stuff" structure.
 */
int
_lw6ker_map_struct_init (lw6ker_map_struct_t * map_struct,
			 lw6map_level_t * level, lw6sys_progress_t * progress)
{
  int ret = 0;
  draft_zones_t *draft_zones;
  int progress_i;
  int progress_n;
  lw6sys_progress_t progress1;
  lw6sys_progress_t progress2;
  lw6sys_progress_t progress3;
  lw6sys_progress_t progress4;

  lw6sys_progress_begin (progress);
  lw6sys_progress_split4 (&progress1, &progress2, &progress3, &progress4,
			  progress);

  memset (map_struct, 0, sizeof (lw6ker_map_struct_t));

  draft_zones = draft_zones_new (level, &progress1);
  if (draft_zones)
    {
      int32_t i;

      lw6sys_progress_begin (&progress2);
      ret = 1;

      i = 1;
      progress_n = 0;
      while (i <= level->param.rules.max_zone_size / 2)
	{
	  progress_n++;
	  i *= 2;
	}

      i = 1;
      progress_i = 0;
      while (i <= level->param.rules.max_zone_size / 2
	     && draft_zones_group (level, draft_zones, i))
	{
	  lw6sys_progress_update (&progress2, 0, progress_n, progress_i++);
	  i *= 2;
	}
      lw6sys_progress_end (&progress2);

      ret = ret
	&& draft_zones_to_map_struct (map_struct, draft_zones, &progress3);

      draft_zones_free (draft_zones);

      ret = ret && init_slots (map_struct, level, &progress4);

      lw6ker_map_struct_sanity_check (map_struct);

      if (ret)
	{
	  lw6sys_log (LW6SYS_LOG_INFO,
		      _
		      ("map struct created %dx%dx%d, %d zones, %1.1f%% compression"),
		      map_struct->shape.w, map_struct->shape.h,
		      map_struct->shape.d, map_struct->nb_zones,
		      _lw6ker_map_struct_get_compression (map_struct) *
		      100.0f);
	}
    }

  lw6sys_progress_end (progress);

  return ret;
}

void
_lw6ker_map_struct_clear (lw6ker_map_struct_t * map_struct)
{
  if (map_struct)
    {
      if (map_struct->zones)
	{
	  LW6SYS_FREE (map_struct->zones);
	}
      if (map_struct->slots)
	{
	  LW6SYS_FREE (map_struct->slots);
	}
      memset (map_struct, 0, sizeof (lw6ker_map_struct_t));
    }
}

void
lw6ker_map_struct_find_free_slot_near (lw6ker_map_struct_t * map_struct,
				       lw6sys_xyz_t * there,
				       lw6sys_xyz_t here)
{
  int found = 0;

  there->x = there->y = there->z = 0;

  if (lw6sys_shape_check_pos (&(map_struct->shape), &here))
    {
      if (lw6ker_map_struct_get_zone_id (map_struct, here.x, here.y, here.z)
	  >= 0)
	{
	  *there = here;
	  found = 1;
	}
      else
	{
	  int32_t angle, radius;
	  int32_t max_radius, max_angle;
	  lw6sys_xyz_t test;

	  max_radius = map_struct->shape.w + map_struct->shape.h;	// +, not *
	  for (radius = 0; radius < max_radius && !found; ++radius)
	    {
	      max_angle = radius * M_PI * 2;
	      for (angle = 0; angle < max_angle && !found; ++angle)
		{
		  test.x =
		    here.x +
		    (lw6ker_cos ((angle * LW6KER_TRIGO_2PI) / max_angle) *
		     radius) / LW6KER_TRIGO_RADIUS;
		  test.y =
		    here.y -
		    (lw6ker_sin ((angle * LW6KER_TRIGO_2PI) / max_angle) *
		     radius) / LW6KER_TRIGO_RADIUS;
		  test.z = 0;
		  if (lw6sys_shape_check_pos (&map_struct->shape, &test))
		    {
		      for (test.z = 0; test.z < map_struct->shape.d; ++test.z)
			{
			  if (lw6ker_map_struct_get_zone_id
			      (map_struct, test.x, test.y, test.z) >= 0)
			    {
			      found = 1;
			      there->x = test.x;
			      there->y = test.y;
			    }
			}
		    }
		}
	    }
	}
    }

  if (!found)
    {
      lw6sys_log (LW6SYS_LOG_WARNING,
		  _("unable to find a free zone near %dx%d"), here.x, here.y);
    }
}

int
lw6ker_map_struct_sanity_check (lw6ker_map_struct_t * map_struct)
{
  int ret = 1;
  int x, y, z, w, h, d;
  int shape_surface, nb_zones, nb_slots;
  int zone_id;
  int i, j;
  int z1, z2;

  w = map_struct->shape.w;
  h = map_struct->shape.h;
  d = map_struct->shape.d;
  shape_surface = map_struct->shape_surface;
  nb_zones = map_struct->nb_zones;
  nb_slots = map_struct->nb_slots;

  if (shape_surface != w * h)
    {
      lw6sys_log (LW6SYS_LOG_WARNING,
		  _("inconsistent shape_surface=%d shape=%dx%d"),
		  shape_surface, w, h);
      ret = 0;
    }

  if (nb_slots != w * h * d)
    {
      lw6sys_log (LW6SYS_LOG_WARNING,
		  _("inconsistent nb_slots=%d shape=%dx%dx%d"), nb_slots, w,
		  h, d);
      ret = 0;
    }

  for (z = 0; z < d; z++)
    {
      for (y = 0; y < h; y++)
	{
	  for (x = 0; x < w; x++)
	    {
	      zone_id = lw6ker_map_struct_get_zone_id (map_struct, x, y, z);
	      if (zone_id == -1 || (zone_id >= 0 && zone_id < nb_zones))
		{
		  if (zone_id >= 0)
		    {
		      z1 = map_struct->zones[zone_id].pos.z;
		      for (i = 0; i < LW6KER_NB_DIRS; ++i)
			{
			  j = map_struct->zones[zone_id].link[i];
			  if (j == -1 || (j >= 0 && j < nb_zones))
			    {
			      if (j >= 0)
				{
				  z2 = map_struct->zones[j].pos.z;
				  switch (i)
				    {
				    case LW6KER_DIR_UP:
				      if (!(z2 == z1 - 1 || z2 == d - 1))
					{
					  lw6sys_log (LW6SYS_LOG_WARNING,
						      _
						      ("bad UP link in %d,%d,%d z1=%d z2=%d"),
						      x, y, z, z1, z2);
					  ret = 0;
					}
				      break;
				    case LW6KER_DIR_DOWN:
				      if (!(z2 == z1 + 1 || z2 == 0))
					{
					  lw6sys_log (LW6SYS_LOG_WARNING,
						      _
						      ("bad DOWN link in %d,%d,%d z1=%d z2=%d"),
						      x, y, z, z1, z2);
					  ret = 0;
					}
				      break;
				    default:
				      // no check for other dirs
				      break;
				    }
				}
			    }
			  else
			    {
			      lw6sys_log (LW6SYS_LOG_WARNING,
					  _("bad link %d->%d in %d,%d,%d"), i,
					  j, x, y, z);
			      ret = 0;
			    }
			}
		    }
		}
	      else
		{
		  lw6sys_log (LW6SYS_LOG_WARNING,
			      _("bad zone id %d in %d,%d,%d"), zone_id, x, y,
			      z);
		  ret = 0;
		}
	    }
	}
    }

  return ret;
}
